Mountaneer = {
	['debug']			= false,	-- false in distribution
	['availableMounts'] = {
		['swimming']		= {},
		['special']			= {},
		['land-slow']		= {},
		['land-fast']	 	= {},
		['flying-slow']		= {},
		['flying-fast']		= {},
		['flying-extreme']	= {}
	},
	['mountDatabase']	= {
		['swimming']		= {},
		['special']			= {},
		['land-slow']		= {},
		['land-fast']		= {},
		['flying-slow']		= {},
		['flying-fast']		= {},
		['flying-extreme']	= {}
	},
	['helper']				= {},	-- helper functions
	['frame']				= CreateFrame('Frame')
};

function Mountaneer:getSwimmingMount()
end;

function Mountaneer:getFlyingMount(speed, name)
	if speed == nil or not (speed == 'slow' or speed == 'fast' or speed == 'rare' or speed == 'epic' or speed == 'extreme') then
		speed = self:getMaxPossibleSpeed('flying');
	end;
	
	if name == 'random' then
		-- pick a random mount of speed [speed]
		return self:getMountByName('flying', speed, '');
	else
		return self:getMountByName('flying', speed, name);
	end;
end;

function Mountaneer:callMount(companionId)
	CallCompanion('MOUNT', companionId);
	local _, name, spellId = GetCompanionInfo('MOUNT', companionId);
	self:showDebugMessage('Chose mount: '..GetSpellLink(spellId)..'(='..spellId..')');
end;

function Mountaneer:hasExtremeMount()
	return (self.helper:tableSize(self.availableMounts['flying-extreme']) > 0);
end;

function Mountaneer.helper:tableSize(tbl)
	local n=0;
	for _,_ in pairs(tbl) do
		n = n+1;
	end;
	return n;
end;

function Mountaneer:getMaxPossibleSpeed(mountType)
	local _, _, _, skillRank = GetSkillLineInfo(12);	-- 12 = riding skill

	if mountType == 'swimming' then
		return '';
	elseif mountType == 'land' then
		if skillRank < 75 then
			return '';
		elseif skillRank == 75 then
			if self.helper:tableSize(self.availableMounts['land-slow']) > 0 then
				return 'slow';
			else
				return '';
			end;
		else
			self:showDebugMessage('found '..tostring(self.helper:tableSize(self.availableMounts['land-slow']))..' slow mounts AND '..tostring(self.helper:tableSize(self.availableMounts['land-fast']))..' fast!');
			if self.helper:tableSize(self.availableMounts['land-fast']) > 0 then
				return 'fast';
			elseif self.helper:tableSize(self.availableMounts['land-slow']) > 0 then
				return 'slow';
			else
				return '';
			end;
		end;
	elseif mountType == 'flying' then
		if skillRank < 225 then
			return '';
		elseif skillRank == 225 then
			if self.helper:tableSize(self.availableMounts['flying-slow']) > 0 then
				return 'slow';
			else
				return '';
			end;
		elseif skillRank == 300 then
			if self:hasExtremeMount() then
				return 'extreme';
			elseif self.helper:tableSize(self.availableMounts['flying-fast']) > 0 then
				return 'fast';
			else
				return 'slow';
			end;
		end;
	else
		-- just the max based on skill
		if skillRank == 75 or skillRank == 225 then
			return 'slow';
		elseif skillRank == 150 then
			return 'fast';
		elseif skillRank == 300 then
			if self:hasExtremeMount() then
				return 'extreme';
			else
				return 'fast';
			end;
		else
			return '';
		end;
	end;
end;

function Mountaneer:getLandMount(speed, name)
	if speed == nil or not (speed == 'slow' or speed == 'fast' or speed == 'rare' or speed == 'epic') then
		speed = self:getMaxPossibleSpeed('land');
	end;

	return self:getMountByName('land', speed, name);
end;

-- update list on riding skill change or new mounts

function Mountaneer:updateAvailableMounts()
	-- clean out our cache first
	self.availableMounts = {
		['swimming'] = {},
		['special'] = {},
		['land-slow'] = {},
		['land-fast'] = {},
		['flying-slow'] = {},
		['flying-fast'] = {},
		['flying-extreme'] = {}
	};
	
	for i=1,GetNumCompanions('MOUNT') do
		local inserted = false;
		_, creatureName, creatureSpellID, _, _ = GetCompanionInfo('MOUNT', i);
		
		-- match each creatureSpellID against all mounts in the database
		for mType, tbl in pairs(self.mountDatabase) do
			for j, spellId in pairs(tbl) do
				if spellId == creatureSpellID then
					table.insert(self.availableMounts[mType], i, spellId);
					inserted = true;
					self:showDebugMessage('added ['..mType..']: '..creatureName);
					break;
				end;
			end;
		end;
		
		if inserted == false then
			self:showChatMessage('new mount discovered which does not appear in the MountaneerDB: \''..creatureName..'\'('..tostring(creatureSpellID)..'). Please alert the addon author of this so it may be added, please also tell what speed the mount is (eg. 150% flying, 60% land, etc.)');
		end;
	end;
end;

function Mountaneer:getMountByName(mountType, speed, name)
	local list = {};
	if mountType == 'any' then
		tables = {'swimming','special','land-'..speed,'flying-'..speed}
		for i, t in pairs(tables) do
			for companionId, spellId in pairs(self.availableMounts[t]) do 
				table.insert(list, companionId);
			end;
		end;
	else
		for companionId, spellId in pairs(self.availableMounts[mountType..'-'..speed]) do 
			self:showDebugMessage('getMountByName: checking companionID: '..GetSpellLink(spellId));
			table.insert(list, companionId);
		end;
	end;
	
	if name == 'random' then
		-- do a random sort and pick one
		local n = self.helper:tableSize(list);
		if n > 0 then
			local r = random(1, n);
			return list[r];
		else
			return nil;
		end;
	else
		finalNameList = {}
		-- go over the list and figure out what we want exactly
		for _, id in pairs(list) do
			_, mountName, spellId = GetCompanionInfo('MOUNT', id);
			mountName = strlower(mountName);	-- assure case-insensitivity
			
			-- check if the name matches and we can use it
			if string.match(mountName, name) then
				table.insert(finalNameList, id);
				self:showDebugMessage('getMountByName: found companionID: '..GetSpellLink(spellId));
			else
				self:showDebugMessage('getMountByName: '..GetSpellLink(spellId)..' <> '..mountName);
			end;
		end;
		local n = self.helper:tableSize(finalNameList);
		if n > 0 then
			-- pick one from the list randomly. If only 1 item, random(1,1) will return 1, so that should work too
			local r = random(1, n);
			_,_,spellId = GetCompanionInfo('MOUNT', finalNameList[r]);
			self:showDebugMessage('chose: '..GetSpellLink(spellId)..' ('..tostring(r)..'/'..self.helper:tableSize(finalNameList)..')');
			return finalNameList[r];
		else
			self:showDebugMessage('no mounts matching query found');
			return nil;
		end;
	end;
end;

function Mountaneer:getMaxMountSearchDetails()
	local currZone = GetRealZoneText();	-- we need to check the current zone against all zones to figure out the continent, and thus if we can fly or not.
	
	-- make a cache for each zone:continent combo
	local zones = {};
	zones['Kalimdor']			= {GetMapZones(1)};
	zones['Eastern Kingdoms']	= {GetMapZones(2)};
	zones['Outlands']			= {GetMapZones(3)};
	zones['Northrend']			= {GetMapZones(4)};
	
	continent = nil;
	for k,zone in pairs(zones['Kalimdor']) do
		if currZone == zone then
			continent = 'vanilla';
			break;
		end;
	end;
	for k,zone in pairs(zones['Eastern Kingdoms']) do
		if currZone == zone then
			continent = 'vanilla';
			break;
		end;
	end;
	for k,zone in pairs(zones['Outlands']) do
		if currZone == zone then
			continent = 'TBC';
			break;
		end;
	end;
	for k,zone in pairs(zones['Northrend']) do
		if currZone == zone then
			continent = 'WotLK';
			break;
		end;
	end;
	
	self:showDebugMessage('You are in a '..tostring(continent)..' zone');
	
	local mountType, speed = nil, nil;
	
	if continent == 'vanilla' then
		-- land mounts only
		mountType = 'land';
		speed = self:getMaxPossibleSpeed('land');
	elseif continent == 'TBC' then
		-- can be either, based on skill
		local _, _, _, skillRank = GetSkillLineInfo(12);	-- 12 = riding skill
		if skillRank > 150 then
			speed = self:getMaxPossibleSpeed('flying');
			if speed == '' then
				speed = self:getMaxPossibleSpeed('flying');
				mountType = 'land';
			else
				mountType = 'flying';
			end;
		else
			mountType = 'land';
			speed = self:getMaxPossibleSpeed('land');
		end;
		
	elseif continent == 'WotLK' then
		-- can be either, based on skill AND coldweather flying AND location! (dalaran!)
		
		-- 1. Zone
		if currZone == 'Dalaran' then
			self:showDebugMessage('Special location: Dalaran');
			-- can only fly from krassus landing
			local sub = GetSubZoneText();
			if sub == "Krasus' Landing" then
				-- only place you can fly
				mountType = 'flying';
				self:showDebugMessage('In Dalaran: Krassus. Can fly here');
			else
				mountType = 'land';
				self:showDebugMessage('In Dalaran: NOT Krassus. Can NOT fly here');
			end;
		elseif currZone == 'Wintergrasp' then
			mountType = 'land';
			self:showDebugMessage('In Wintergrasp. Can NOT fly here');
		else
			-- just fly
			mountType = 'flying';
		end;
		
		-- 2. cold weather flying?
		if not IsUsableSpell(GetSpellInfo(54197)) then
			-- 54197 = spellId for Cold Weather Flying
			mountType = 'land';
			self:showDebugMessage('You do not have Cold Weather Flying!');
		end;
		
		-- 3. enough skill ?
		local _, _, _, skillRank = GetSkillLineInfo(12);	-- 12 = riding skill
		if skillRank < 225 then
			mountType = 'land';	-- regardless of location, we can only use landmounts anyway
		end;
		speed = self:getMaxPossibleSpeed(mountType);
	else
		mountType = '';
		speed = '';
	end;
	
	return mountType, speed;
end;

function Mountaneer:showChatHelp()
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ff00Mountaneer - easy chat commands for extensive mount collections|r')
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ff00Usage: |r|cff00ffff/mount [actionType] [mountType] [speed] [name]|r');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ff00/mount|r');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ffff        [actionType]:');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ffff           max - mounts best available mount possible. All other options hereafter are ignored.');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ffff           list - shows list of mounts matching the search instead of auto-mounting');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ffff           [none] - skipping this option will auto-mount a random mount from the list');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ffff        [mountType]:');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ffff           special - adds only special mounts to the list which do not increase speed. Currently only the Riding Turtle');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ffff           swimming - adds only swimming mounts to the list. Currently only the Sea Turtle');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ffff           land - adds only land mounts to the list.');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ffff           flying - adds only flying mounts to the list.');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ffff           [none] - skipping this option will add all mount-types to the list.');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ffff        [speed]:');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ffff           slow - adds only slow mounts to the list (+60% for land mounts, +150% for flying');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ffff           rare - alias for slow');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ffff           fast - adds only fast mounts to the list (+100% for land mounts, +280% for flying');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ffff           epic - alias for fast.');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ffff           extreme - adds only extremely fast mounts to the list (+310% flying mounts).');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ffff           [none] - skipping this option will add mounts of all speeds to the list.');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ffff        [name]:');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ffff           random - this special option will choose a random mount from the list.');
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ffff           [name] - by entering a (partial) name of a mount here, you can narrow down the search');
	DEFAULT_CHAT_FRAME:AddMessage();
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ff00example:|r|cffffff00/mount land fast saber|r - will mount a random saber which runs at +100% speed. eg. a Swift Frostsaber');
end;

function Mountaneer:load()
	self.frame:UnregisterEvent('PLAYER_LOGIN');
	
	self.frame:RegisterEvent('COMPANION_UPDATE');	
	self.frame:SetScript('OnEvent', function() self:updateAvailableMounts() end);
	
	self:showChatMessage('succesfully loaded');
end;


SLASH_MOUNTANEER1 = '/mount';
SlashCmdList['MOUNTANEER'] = function(input)
	--[[
		-- input may contain:
		swim (=summon turtle [if any])
		land [slow/fast | rare/epic] [name]
		flying [slow/fast/extreme | rare/epic/special] [name]
		[name] (partial or full. Use quotes for exact)
		
		max (= best mount available here. Special support for Dalaran and WinterGrasp)	-- to be implemented
		
		list [swim|land|flying] [slow|fast|extreme|rare|epic] [name]	-- to be implemented
	]]--
	
	-- hack
	self = Mountaneer;
	
	-- assure case-insensitivity
	input = strlower(input);
	
	self:showDebugMessage('got: '..tostring(input)..' via /mount');
	
	local mountType, speed, name, asList = nil, nil, {}, false;
	local skip = false;
	for m in string.gmatch(input, '[^ ]+') do
		if (not mountType and not speed and m == 'init') then
			self:updateAvailableMounts();
			return;
		end;
		
		if (not mountType and not speed) then
			-- special case
			if m == 'max' then
				mountType, speed = self:getMaxMountSearchDetails();
				name = {'random'};
				break;
			elseif m == 'list' then
				asList = true;
				skip = true;	-- skip one run
			elseif m == 'help' then
				self:showChatHelp();
				return;
			end;
		end;
		
		if (not skip and (not mountType or not speed)) then
			if not mountType then
				if (m == 'swim' or m == 'land' or m == 'flying') then
					mountType = m;
				else
					mountType = 'any';
					
					-- apparently we're going for a direct name approach
					table.insert(name, m);
				end;
			else
				if (m == 'slow' or m == 'rare') and speed == nil then
					speed = 'slow';
				elseif (m == 'fast' or m == 'epic') and speed == nil then
					speed = 'fast';
				elseif m == 'extreme' and speed == nil then
					speed = 'extreme';
				else
					speed = self.getMaxPossibleSpeed(mountType);
					-- apparently we're at the name now
					table.insert(name, m);
				end;
			end;
		elseif not skip then
			table.insert(name, m);
		else
			skip = false;
		end;
	end;
	if mountType == nil then
		self:showChatHelp();
		return;
	end;
	
	-- make sure we get a mount name, default to special:random
	if name == nil then
		name = {'random'};
	end;
	
	-- concat the name table, this makes sure we support mount names which contain spaces (eg. Swift Stormsaber)
	local nName = nil;
	for i, val in pairs(name) do
		if(nName == nil) then
			nName = '';
		else
			nName = nName..' ';
		end;
		nName = nName..val;
	end;
	
	self:showDebugMessage('Calling ['..mountType..'('..speed..')] "'..nName..'"');
	self:showChatMessage('Looking for a '..speed..' '..mountType..' mount, '..nName);
	
	if asList then
		-- show a list of possible mounts
		self:showChatMessage('List is currently not implemented');	-- TO BE IMPLEMENTED
	else
		-- auto-summon a random mount from the list
		local mount = nil;
		if(mountType == 'swim') then
			mount = self:getSwimmingMount();	
		elseif(mountType == 'land') then
			mount = self:getLandMount(speed, nName);
		elseif(mountType == 'flying') then
			mount = self:getFlyingMount(speed, nName);
		else
			mount = self:getMountByName('any', speed, nName);
		end;
		
		-- check if we actually have any mounts that match the query at all
		if mount == nil then
			self:showChatMessage('No mount matching your query has been found.');
		else
			self:callMount(mount);
		end;
	end;	
end;

function Mountaneer:showChatMessage(msg)
	DEFAULT_CHAT_FRAME:AddMessage('|cff00ff00Mountaneer:|r '..msg);
end;

function Mountaneer:showDebugMessage(msg)
	if self.debug then
		DEFAULT_CHAT_FRAME:AddMessage('|cff00ff00Mountaneer DEBUG:|r |cffff0000'..msg..'|r');
	end;
end;

Mountaneer.frame:RegisterEvent('PLAYER_LOGIN');
Mountaneer.frame:SetScript('OnEvent', function(self, event, ...) Mountaneer:load() end);